/****************************************************************************
 *
 * type1cid.c
 *
 *   FreeType OpenType driver component (body only).
 *
 * Copyright (C) 1996-2025 by
 * David Turner, Robert Wilhelm, and Werner Lemberg.
 *
 * This file is part of the FreeType project, and may only be used,
 * modified, and distributed under the terms of the FreeType project
 * license, LICENSE.TXT.  By continuing to use, modify, or distribute
 * this file you indicate that you have read the license and
 * understand and accept it fully.
 *
 */


#define FT_MAKE_OPTION_SINGLE_OBJECT

#include "cidgload.c"
#include "cidload.c"
#include "cidobjs.c"
#include "cidparse.c"
#include "cidriver.c"


/* END */
