#version 100
precision mediump float;
precision highp int;

vec2 _19;

void main()
{
    highp vec2 _30;
    for (int spvDummy15 = 0; spvDummy15 < 1; spvDummy15++)
    {
        if (gl_FragCoord.x != gl_FragCoord.x)
        {
            _30 = _19;
            break;
        }
        highp vec2 _29;
        _29.y = _19.y;
        _30 = _29;
        break;
    }
    gl_FragData[0] = vec4(_30, 1.0, 1.0);
}

